/*
 *                        Lamma Regione Toscana
 *
 * RCS      : $Id: read_grib.c,v 1.1.1.1 2004/09/07 13:31:55 ocean Exp $
 * Author   : Graziano Giuliani
 * Project  : WW3 forecast system
 * Date     : Fri May 30 15:43:52 CEST 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <time.h>
#include <grib.h>
#include <ebisuzaki.h>

static size_t pos    = 0;
static size_t size   = 0;
static size_t limit  = 0;
static size_t reclen = 0;
static unsigned char *data    = NULL;
static unsigned char *message = NULL;
static unsigned char *pds     = NULL;
static unsigned char *gds     = NULL;
static unsigned char *bms     = NULL;
static unsigned char *bds     = NULL;
static int fd        = -1;
static int verbose   = 0;

void cname(char *src, char *dest, int maxlen);

void open_grib(char *filename, int *istat, int fsize)
{
  char fname[256];
  struct stat finfo;
  
  cname(filename, fname, fsize);

  if (stat(fname, &finfo) < 0)
  {
    fprintf(stderr, "Error opening input file %s.\n", fname);
    *istat = -1;
    return;
  }

  size = finfo.st_size;
  limit = size - 8;

  fd = open(fname, O_RDONLY);
  if (fd < 0)
  {
    fprintf(stderr, "Error opening input file %s.\n", fname);
    *istat = -1;
    return;
  }

  data = mmap(0, size, PROT_READ, MAP_PRIVATE, fd, 0);
  if (data == MAP_FAILED)
  {
    fprintf(stderr, "Error reading input file %s.\n", fname);
    *istat = -2;
    return;
  }

  *istat = 0;
  return;
}

void read_record(int *istat)
{
  unsigned char *pnt;

  if (message != NULL) free(message);
  message = NULL;
  pds     = NULL;
  gds     = NULL;
  bms     = NULL;
  bds     = NULL;
  reclen = 0;

  if (pos >= limit)
  {
    *istat = 1;
    return;
  }

  while ((data[pos+0] != 'G'  &&
          data[pos+1] != 'R'  &&
          data[pos+2] != 'I'  &&
          data[pos+3] != 'B') &&
         pos < limit)
  {
    pos ++;
  }

  if (pos >= limit)
  {
    *istat = 1;
    return;
  }

  reclen = (data[pos+4]<<16) + (data[pos+5]<<8) + data[pos+6];

  if (verbose)
    fprintf(stdout, "GRIB message found at pos %d: len %d\n", pos, reclen);

  if ((pos + reclen) > size)
  {
    fprintf(stderr, "Incomplete GRIB message.\n");
    *istat = -1;
    return;
  }

  if (data[pos+7] > 1)
  {
    fprintf(stderr, "GRIB message version is %d.\n", data[pos+7]);
    *istat = -2;
    return;
  }
 
  message = malloc(reclen);
  if (message == NULL)
  {
    fprintf(stderr, "Memory allocation error of %d bytes.\n", reclen);
    *istat = -3;
    return;
  }

  memcpy(message, data+pos, reclen);
  pos = pos + reclen;

  pds = message + 8;
  pnt = pds + PDS_LEN(pds);
  if (PDS_HAS_GDS(pds))
  {
    gds = pnt;
    pnt += GDS_LEN(gds);
  }
  if (PDS_HAS_BMS(pds))
  {
    bms = pnt;
    pnt += BMS_LEN(bms);
  }
  bds = pnt;
  pnt += BDS_LEN(bds);

  if ((size_t) (pnt-message+4) != reclen)
  {
    fprintf(stderr, "Len of grib message is inconsistent.\n");
    *istat = -2;
    return;
  }

  *istat = 0;
  return;
}

void close_grib( )
{
  munmap(data, size);
  close(fd);
  pos     = 0;
  size    = 0;
  limit   = 0;
  reclen  = 0;
  data    = NULL;
  message = NULL;
  pds     = NULL;
  gds     = NULL;
  bms     = NULL;
  bds     = NULL;
  fd      = -1;
  return;
}

void print_pds_hex( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  fprintf(stdout, "PDS = ");
  for (i = 0; i < PDS_LEN(pds); i ++)
    fprintf(stdout, "%2X ", pds[i]);
  fprintf(stdout, "\n");

  return;
}

void print_pds_dec( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  fprintf(stdout, "PDS = ");
  for (i = 0; i < PDS_LEN(pds); i ++)
    fprintf(stdout, "%3d ", pds[i]);
  fprintf(stdout, "\n");

  return;
}

void print_gds_hex( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  if (PDS_HAS_GDS(pds))
  {
    fprintf(stdout, "GDS = ");
    for (i = 0; i < GDS_LEN(gds); i ++)
      fprintf(stdout, "%2X ", gds[i]);
    fprintf(stdout, "\n");
  }
  else
    fprintf(stdout, "GDS not present in message.\n");

  return;
}

void print_gds_dec( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  if (PDS_HAS_GDS(pds))
  {
    fprintf(stdout, "GDS = ");
    for (i = 0; i < GDS_LEN(gds); i ++)
      fprintf(stdout, "%3d ", gds[i]);
    fprintf(stdout, "\n");
  }
  else
    fprintf(stdout, "GDS not present in message.\n");

  return;
}

void print_bms_hex( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  if (PDS_HAS_BMS(pds))
  {
    fprintf(stdout, "BMS = ");
    for (i = 0; i < BMS_LEN(bms); i ++)
      fprintf(stdout, "%2X ", bms[i]);
    fprintf(stdout, "\n");
  }
  else
    fprintf(stdout, "BMS not present in message.\n");

  return;
}

void print_bms_dec( )
{
  int i;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  if (PDS_HAS_BMS(pds))
  {
    fprintf(stdout, "BMS = ");
    for (i = 0; i < BMS_LEN(bms); i ++)
      fprintf(stdout, "%2X ", bms[i]);
    fprintf(stdout, "\n");
  }
  else
    fprintf(stdout, "BMS not present in message.\n");

  return;
}

int get_pds_value(int *index)
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }

  if (*index < 0 || *index > PDS_LEN(pds))
  {
    fprintf(stderr, "Requested PDS index out of range.\n");
    return -1;
  }
  return (int) *(pds+*index);
}

int is_gds_present( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_HAS_GDS(pds));
}

int get_gds_value(int *index)
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }

  if (!PDS_HAS_GDS(pds))
  {
    fprintf(stderr, "Message without GDS record.\n");
    return -1;
  }
  if (*index < 0 || *index > GDS_LEN(gds))
  {
    fprintf(stderr, "Requested GDS index out of range.\n");
    return -1;
  }
  return (int) *(gds+*index);
}

int is_bms_present( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_HAS_BMS(pds));
}

int get_bms_value(int *index)
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }

  if (!PDS_HAS_BMS(pds))
  {
    fprintf(stderr, "Message without BMS record.\n");
    return -1;
  }
  if (*index < 0 || *index > BMS_LEN(bms))
  {
    fprintf(stderr, "Requested BMS index out of range.\n");
    return -1;
  }
  return (int) *(bms+*index);
}

void get_grid_size(long *nx, long *ny, long *nxny)
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  *ny = 1;
  if (gds != NULL)
    GDS_grid(gds, bds, nx, ny, nxny);
  else if (bms != NULL)
    *nxny = *nx = BMS_nxny(bms);
  else
  {
    if (BDS_NumBits(bds) == 0)
      *nxny = *nx = 1;
    else
      *nxny = *nx = BDS_NValues(bds);
  }
  return;
}

void get_variable_name(char *vname, int vsize)
{
  size_t as;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  memset(vname, 32, vsize);
  as = ((size_t) vsize < strlen((Parm_Table(pds) + PDS_PARAM(pds))->name)) ? 
        (size_t) vsize : strlen((Parm_Table(pds) + PDS_PARAM(pds))->name);
  memcpy(vname, (Parm_Table(pds) + PDS_PARAM(pds))->name, as);
  return;
}

void get_variable_desc(char *vdesc, int vsize)
{
  size_t as;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  memset(vdesc, 32, vsize);
  as = ((size_t) vsize < strlen((Parm_Table(pds) + PDS_PARAM(pds))->comment)) ? 
        (size_t) vsize : strlen((Parm_Table(pds) + PDS_PARAM(pds))->comment);
  memcpy(vdesc, (Parm_Table(pds) + PDS_PARAM(pds))->comment, as);
  return;
}

int is_level_pressure( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_KPDS6(pds) == 100);
}

int is_level_z( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_KPDS6(pds) == 105);
}

int is_surface( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_KPDS6(pds) == 1);
}

int is_mean_sea_level( )
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  return (PDS_KPDS6(pds) == 102);
}

void get_level_value(float *val)
{
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  *val = ((float) PDS_KPDS7(pds));
  return;
}

void get_level_desc(char *ldesc, int vsize)
{
  size_t as;
  char *tmp;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  tmp = levels(pds);
  memset(ldesc, 32, vsize);
  as = ((size_t) vsize) < strlen(tmp) ? (size_t) vsize : strlen(tmp);
  memcpy(ldesc, tmp, as);
  return;
}

void get_reftime(char *stime, int vsize)
{
  size_t as;
  char tmp[15];

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  memset(stime, 32, vsize);
  sprintf(tmp, "%04d%02d%02d%02d%02d00", PDS_Year4(pds),
          PDS_Month(pds), PDS_Day(pds), PDS_Hour(pds), PDS_Minute(pds));
  as = ((size_t) vsize < strlen(tmp)) ?  (size_t) vsize : strlen(tmp);
  memcpy(stime, tmp, as);
  return;
}

void get_forecast_time(int *rincr)
{
  int tr;
  int dtime;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  dtime = PDS_P1(pds);
  tr = PDS_TimeRange(pds);
  if (tr == 10) dtime = PDS_P1(pds) * 256 + PDS_P2(pds);
  if (tr > 1 && tr < 6 ) dtime = PDS_P2(pds);

  if (PDS_ForecastTimeUnit(pds) == MINUTE)
    *rincr = dtime * 60;
  else if (PDS_ForecastTimeUnit(pds) == HOUR)
    *rincr = dtime * 3600;
  else if (PDS_ForecastTimeUnit(pds) == DAY)
    *rincr = dtime * 86400;
  else if (PDS_ForecastTimeUnit(pds) == MONTH)
    *rincr = dtime * 2592000;
  else if (PDS_ForecastTimeUnit(pds) == HOURS3)
    *rincr = dtime * 10800;
  else if (PDS_ForecastTimeUnit(pds) == HOURS6)
    *rincr = dtime * 21600;
  else if (PDS_ForecastTimeUnit(pds) == HOURS12)
    *rincr = dtime * 43200;
  else if (PDS_ForecastTimeUnit(pds) == SECOND)
    *rincr = dtime;
  else
  {
    fprintf(stdout, "Cannot determine forecast time!!!!");
    *rincr = -1;
    return;
  }
  return;
}

void get_datetime(int *year, int *month, int *day, int *hour, int *minute)
{
  struct tm itm, *xtm;
  long incr;
  time_t ttm;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }

  *year = PDS_Year4(pds);
  *month = PDS_Month(pds);
  *day = PDS_Day(pds);
  *hour = PDS_Hour(pds);
  *minute = PDS_Minute(pds);

  if (PDS_P2(pds))
  {
    memset(&itm, 0, sizeof(struct tm));
    itm.tm_year = PDS_Year4(pds) - 1900;
    itm.tm_mon  = PDS_Month(pds) - 1;
    itm.tm_mday = PDS_Day(pds);
    itm.tm_hour = PDS_Hour(pds);
    itm.tm_min  = PDS_Minute(pds);

    ttm = mktime(&itm);

    if (PDS_ForecastTimeUnit(pds) == MINUTE)
      incr = PDS_P2(pds) * 60;
    else if (PDS_ForecastTimeUnit(pds) == HOUR)
      incr = PDS_P2(pds) * 3600;
    else if (PDS_ForecastTimeUnit(pds) == DAY)
      incr = PDS_P2(pds) * 86400;
    else if (PDS_ForecastTimeUnit(pds) == MONTH)
      incr = PDS_P2(pds) * 2592000;
    else if (PDS_ForecastTimeUnit(pds) == HOURS3)
      incr = PDS_P2(pds) * 10800;
    else if (PDS_ForecastTimeUnit(pds) == HOURS6)
      incr = PDS_P2(pds) * 21600;
    else if (PDS_ForecastTimeUnit(pds) == HOURS12)
      incr = PDS_P2(pds) * 43200;
    else if (PDS_ForecastTimeUnit(pds) == SECOND)
      incr = PDS_P2(pds);
    else
    {
      fprintf(stdout, "Cannot determine forecast time!!!!");
      return;
    }
    tzset( );
    ttm = ttm + incr - timezone;
    xtm = gmtime(&ttm);
    *year = xtm->tm_year + 1900;
    *month = xtm->tm_mon + 1;
    *day = xtm->tm_mday;
    *hour = xtm->tm_hour;
    *minute = xtm->tm_min;
    return;
  }

  return;
}

int is_grid_latlon( )
{
  long nx, ny, nxny;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return -1;
  }
  get_grid_size(&nx, &ny, &nxny);
  return (gds && GDS_LatLon(gds) && nx != -1);
}

void get_latlon_grid_info(float *minlat, float *maxlat, float *dlat,
                           float *minlon, float *maxlon, float *dlon)
{
  long nx, ny, nxny;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_grid_size(&nx, &ny, &nxny);
  if (gds && GDS_LatLon(gds) && nx != -1)
  {
    *minlat = 0.001*GDS_LatLon_La2(gds);
    *maxlat = 0.001*GDS_LatLon_La1(gds);
    *minlon = 0.001*GDS_LatLon_Lo2(gds);
    *maxlon = 0.001*GDS_LatLon_Lo1(gds);
    *dlat   = 0.001*GDS_LatLon_dy(gds);
    *dlon   = 0.001*GDS_LatLon_dx(gds);
  }
  else
    fprintf(stderr, "This grid is not Latlon.\n");
  return;
}

void get_data(int *inx, int *iny, float *data)
{
  long nx, ny, nxny;
  int i, j;
  double temp;
  float *vals;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_grid_size(&nx, &ny, &nxny);

  if (*inx < nx || *iny < ny)
  {
    fprintf(stderr, "Needed matrix %ldx%ld\n", ny, nx);
    return;
  }

  vals = malloc(nxny*sizeof(float));
  if (vals == NULL)
  {
    fprintf(stderr, "Memory error in get_data.\n");
    return;
  }

  temp = int_power(10.0, - PDS_DecimalScale(pds));
  BDS_unpack(vals, bds, BMS_bitmap(bms), BDS_NumBits(bds), nxny,
             temp*BDS_RefValue(bds),temp*int_power(2.0, BDS_BinScale(bds)));

  for (j = 0; j < ny; j ++)
    for (i = 0; i < nx; i ++)
      data[i*(*iny)+(*iny)-1-j] = vals[j*nx+i];

  free(vals);
  return;
}

void cname(char *src, char *dest, int maxlen)
{
  char *pnt;
  memcpy(dest, src, maxlen-1);
  dest[maxlen-1] = 0;
  pnt = strrchr(dest, ' ');
  if (pnt == NULL)
    return;
  while (*(pnt-1) == ' ') pnt --;
  dest[pnt-dest] = 0;
  return;
}

void print_grid_size( )
{
  long nx, ny, nxny;
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_grid_size(&nx, &ny, &nxny);
  fprintf(stdout, "Grid size: %ld (%ldx%ld)\n", nxny, ny, nx);
  return;
}

void print_variable_name( )
{
  char vname[64];
  char cvname[64];
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_variable_name(vname, 64);
  cname(vname, cvname, 64);
  fprintf(stdout, "Variable name: %s\n", cvname);
  return;
}

void print_variable_desc( )
{
  char vdesc[128];
  char cvdesc[128];
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_variable_desc(vdesc, 128);
  cname(vdesc, cvdesc, 128);
  fprintf(stdout, "Variable Description: %s\n", cvdesc);
  return;
}

void print_datetime( )
{
  int year, month, day, hour, minute;
  char reftime[32];
  char creftime[32];
  int fcst = 0;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_datetime(&year, &month, &day, &hour, &minute);
  get_reftime(reftime, 32);
  cname(reftime, creftime, 32);
  get_forecast_time(&fcst);
  fprintf(stdout, "Date: %04d-%02d-%02d %02d:%02d:00 UTC (%s + %d sec)\n",
          year, month, day, hour, minute, creftime, fcst );
  return;
}

void print_grid_info( )
{
  long nx, ny, nxny;
  int iy, i, col, pl;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_grid_size(&nx, &ny, &nxny);
  fprintf(stdout, "Grid information:\n");
  if (gds && GDS_LatLon(gds) && nx != -1) 
    fprintf(stdout, "\tlatlon: lat  %f to %f by %f  nxny %ld\n"
                    "\t        long %f to %f by %f, (%ld x %ld) scan %d\n"
                    "\t        mode %d bdsgrid %d\n",
            0.001*GDS_LatLon_La1(gds), 0.001*GDS_LatLon_La2(gds),
            0.001*GDS_LatLon_dy(gds), nxny, 0.001*GDS_LatLon_Lo1(gds),
            0.001*GDS_LatLon_Lo2(gds), 0.001*GDS_LatLon_dx(gds),
            nx, ny, GDS_LatLon_scan(gds), GDS_LatLon_mode(gds),
            BDS_Grid(bds));
  else if (gds && GDS_LatLon(gds) && nx == -1)
  {
    fprintf(stdout, "thinned latlon: lat %f to %f by %f nxny %ld\n"
                    "long %f to %f, %ld grid pts   (%ld x %ld) scan %d"
                    " mode %d bdsgrid %d\n",
            0.001*GDS_LatLon_La1(gds), 0.001*GDS_LatLon_La2(gds),
            0.001*GDS_LatLon_dy(gds), nxny, 0.001*GDS_LatLon_Lo1(gds),
            0.001*GDS_LatLon_Lo2(gds),
            nxny, nx, ny, GDS_LatLon_scan(gds), GDS_LatLon_mode(gds),
            BDS_Grid(bds));
    iy = GDS_LatLon_ny(gds);
    iy = (iy + 1) / 2;
    iy = GDS_LatLon_ny(gds);
    if ((pl = GDS_PL(gds)) == -1)
    {
      fprintf(stderr,"Thinned Latlon grid inconsistent.\n");
      return;
    }
    for (col = i = 0; i < iy; i++)
    {
      if (col == 0) printf("   ");
      fprintf(stdout, "%5d", (gds[pl+i*2] << 8) + gds[pl+i*2+1]);
      col++;
      if (col == 15)
      {
        col = 0;
        fprintf(stdout, "\n");
      }
    }
    if (col != 0) fprintf(stdout, "\n");
  }
  else if (gds && GDS_Gaussian(gds) && nx != -1)
    fprintf(stdout, "gaussian: lat  %f to %f\n"
                    "long %f to %f by %f, (%ld x %ld) scan %d"
                    " mode %d bdsgrid %d\n",
            0.001*GDS_LatLon_La1(gds), 0.001*GDS_LatLon_La2(gds),
            0.001*GDS_LatLon_Lo1(gds), 0.001*GDS_LatLon_Lo2(gds), 
            0.001*GDS_LatLon_dx(gds), nx, ny,
            GDS_LatLon_scan(gds), GDS_LatLon_mode(gds),
            BDS_Grid(bds));
  else if (gds && GDS_Gaussian(gds) && nx == -1)
  {
    fprintf(stdout, "thinned gaussian: lat  %f to %f\n"
                    "lon %f   %ld grid pts   (%ld x %ld) scan %d"
                    " mode %d bdsgrid %d  nlat:\n",
            0.001*GDS_LatLon_La1(gds), 0.001*GDS_LatLon_La2(gds),
            0.001*GDS_LatLon_Lo1(gds),
            nxny, nx, ny, GDS_LatLon_scan(gds), GDS_LatLon_mode(gds),
            BDS_Grid(bds));
    iy = GDS_LatLon_ny(gds);
    iy = (iy + 1) / 2;
    iy = GDS_LatLon_ny(gds);
    if ((pl = GDS_PL(gds)) == -1)
    {
      fprintf(stderr,"Thinned Latlon grid inconsistent.\n");
      return;
    }
    for (col = i = 0; i < iy; i++)
    {
      if (col == 0) printf("   ");
      fprintf(stdout, "%5d", (gds[pl+i*2] << 8) + gds[pl+i*2+1]);
      col++;
      if (col == 15)
      {
        col = 0;
        fprintf(stdout, "\n");
      }
    }
    if (col != 0) fprintf(stdout, "\n");
  }
  else if (gds && GDS_Polar(gds))
    fprintf(stdout, "polar stereo: Lat1 %f Long1 %f Orient %f\n"
                    "%s pole (%ld x %ld) Dx %d Dy %d scan %d mode %d\n",
            0.001*GDS_Polar_La1(gds),0.001*GDS_Polar_Lo1(gds),
            0.001*GDS_Polar_Lov(gds),
            GDS_Polar_pole(gds) == 0 ? "north" : "south", nx,ny,
            GDS_Polar_Dx(gds),GDS_Polar_Dy(gds),
            GDS_Polar_scan(gds), GDS_Polar_mode(gds));
  else if (gds && GDS_Lambert(gds))
    fprintf(stdout, "Lambert Conf: Lat1 %f Lon1 %f Lov %f\n"
                    "Latin1 %f Latin2 %f LatSP %f LonSP %f\n"
                    "%s (%d x %d) Dx %f Dy %f scan %d mode %d\n",
            0.001*GDS_Lambert_La1(gds),0.001*GDS_Lambert_Lo1(gds),
            0.001*GDS_Lambert_Lov(gds),
            0.001*GDS_Lambert_Latin1(gds), 0.001*GDS_Lambert_Latin2(gds),
            0.001*GDS_Lambert_LatSP(gds), 0.001*GDS_Lambert_LonSP(gds),
            GDS_Lambert_NP(gds) ? "North Pole": "South Pole",
            GDS_Lambert_nx(gds), GDS_Lambert_ny(gds),
            0.001*GDS_Lambert_dx(gds), 0.001*GDS_Lambert_dy(gds),
            GDS_Lambert_scan(gds), GDS_Lambert_mode(gds));
  else if (gds && GDS_Mercator(gds))
    fprintf(stdout, "Mercator: lat  %f to %f by %f km  nxny %ld\n"
                    "long %f to %f by %f km, (%ld x %ld) scan %d"
                    " mode %d Latin %f bdsgrid %d\n",
            0.001*GDS_Merc_La1(gds), 0.001*GDS_Merc_La2(gds),
            0.001*GDS_Merc_dy(gds), nxny, 0.001*GDS_Merc_Lo1(gds),
            0.001*GDS_Merc_Lo2(gds), 0.001*GDS_Merc_dx(gds),
            nx, ny, GDS_Merc_scan(gds), GDS_Merc_mode(gds), 
            0.001*GDS_Merc_Latin(gds), BDS_Grid(bds));
   else if (gds && GDS_ssEgrid(gds))
     fprintf(stdout, "Semi-staggered Arakawa E-Grid: lat0 %f lon0 %f nxny %d\n"
                     "dLat %f dLon %f (%d x %d) scan %d mode %d\n",
             0.001*GDS_ssEgrid_La1(gds), 0.001*GDS_ssEgrid_Lo1(gds), 
             GDS_ssEgrid_n(gds)*GDS_ssEgrid_n_dum(gds), 
             0.001*GDS_ssEgrid_dj(gds), 0.001*GDS_ssEgrid_di(gds), 
             GDS_ssEgrid_Lo2(gds), GDS_ssEgrid_La2(gds),
             GDS_ssEgrid_scan(gds), GDS_ssEgrid_mode(gds));
  else if (gds && GDS_ss2dEgrid(gds))
    fprintf(stdout, "Semi-stagg. Arakawa E-Grid (2D): lat0 %f lon0 %f nxny %d\n"
                    "dLat %f dLon %f (tlm0d %f tph0d %f) scan %d mode %d\n",
            0.001*GDS_ss2dEgrid_La1(gds), 0.001*GDS_ss2dEgrid_Lo1(gds),
            GDS_ss2dEgrid_nx(gds)*GDS_ss2dEgrid_ny(gds),
            0.001*GDS_ss2dEgrid_dj(gds), 0.001*GDS_ss2dEgrid_di(gds),
            0.001*GDS_ss2dEgrid_Lo2(gds), 0.001*GDS_ss2dEgrid_La2(gds),
            GDS_ss2dEgrid_scan(gds), GDS_ss2dEgrid_mode(gds));
  else if (gds && GDS_fEgrid(gds)) 
    fprintf(stdout, "filled Arakawa E-Grid: lat0 %f lon0 %f nxny %d\n"
                    "dLat %f dLon %f (%d x %d) scan %d mode %d\n",
            0.001*GDS_fEgrid_La1(gds), 0.001*GDS_fEgrid_Lo1(gds), 
            GDS_fEgrid_n(gds)*GDS_fEgrid_n_dum(gds), 
            0.001*GDS_fEgrid_dj(gds), 0.001*GDS_fEgrid_di(gds), 
            GDS_fEgrid_Lo2(gds), GDS_fEgrid_La2(gds),
            GDS_fEgrid_scan(gds), GDS_fEgrid_mode(gds));
  else if (gds && GDS_RotLL(gds))
    fprintf(stdout, "rotated LatLon grid  lat %f to %f  lon %f to %f\n"
                    "nxny %ld  (%d x %d)  dx %d dy %d  scan %d  mode %d\n"
                    "transform: south pole lat %f lon %f  rot angle %f\n", 
            0.001*GDS_RotLL_La1(gds), 0.001*GDS_RotLL_La2(gds), 
            0.001*GDS_RotLL_Lo1(gds), 0.001*GDS_RotLL_Lo2(gds),
            nxny, GDS_RotLL_nx(gds), GDS_RotLL_ny(gds),
            GDS_RotLL_dx(gds), GDS_RotLL_dy(gds),
            GDS_RotLL_scan(gds), GDS_RotLL_mode(gds),
            0.001*GDS_RotLL_LaSP(gds), 0.001*GDS_RotLL_LoSP(gds),
            GDS_RotLL_RotAng(gds));
  else if (gds && GDS_Gnomonic(gds))
    fprintf(stdout, "Gnomonic grid\n");
  else if (gds && GDS_Harmonic(gds))
    fprintf(stdout, "Harmonic (spectral): pentagonal"
                    " spectral truncation: nj %d nk %d nm %d\n",
            GDS_Harmonic_nj(gds), GDS_Harmonic_nk(gds),
            GDS_Harmonic_nm(gds));
  if (gds && GDS_Harmonic_type(gds) == 1)
    fprintf(stdout, "Associated Legendre polynomials\n");
  else if (gds && GDS_Triangular(gds))
    fprintf(stdout, "Triangular grid:  nd %d ni %d (= 2^%d x 3^%d)\n",
            GDS_Triangular_nd(gds), GDS_Triangular_ni(gds),
            GDS_Triangular_ni2(gds), GDS_Triangular_ni3(gds) );
  return;
}

void print_level( )
{
  char ldesc[64];
  char cldesc[64];
  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_level_desc(ldesc, 64);
  cname(ldesc, cldesc, 64);
  fprintf(stdout, "Level : %s\n", cldesc);
  return;
}

void print_data( )
{
  long nx, ny, nxny;
  int i, j;
  float *vals;
  double temp;
  int lastidx;

  if (pds == NULL)
  {
    fprintf(stderr, "No message has been read in.\n");
    return;
  }
  get_grid_size(&nx, &ny, &nxny);
  vals = malloc(nxny*sizeof(float));
  if (vals == NULL)
  {
    fprintf(stderr, "Memory error in print_data.\n");
    return;
  }

  temp = int_power(10.0, - PDS_DecimalScale(pds));

  BDS_unpack(vals, bds, BMS_bitmap(bms), BDS_NumBits(bds), nxny,
             temp*BDS_RefValue(bds),temp*int_power(2.0, BDS_BinScale(bds)));

  lastidx = (nxny/8*8);
  fprintf(stdout, "########################################################\n");
  for (i = 0; i < lastidx; i=i+8)
  {
    for (j = 0; j < 8; j ++)
      fprintf(stdout, "%10.2f", vals[i+j]);
    fprintf(stdout, "\n");
  }
  for (i = lastidx; i < nxny; i ++)
    fprintf(stdout, "%10.2f", vals[i]);
  fprintf(stdout, "\n");
  fprintf(stdout, "########################################################\n");
  free(vals);
  return;
}

#ifdef TESTME

int main(int argc, char *argv[])
{
  int istat;
  /* char fname[] = "test.grb                               "; */
  char fname[1024];

  memset(fname, 32, 1024);
  memcpy(fname, argv[1], strlen(argv[1]));

  open_grib( fname, &istat, 128);
  if (istat) return -1;

  while (istat == 0)
  {
    read_record(&istat);
    if (istat == 0)
    {
      print_variable_name( );
      print_variable_desc( );
      print_datetime( );
      print_grid_size( );
      print_grid_info( );
      print_level( );
      print_data( );
    }
  }

  close_grib( );

  if (istat < 0)
    return -1;

  fprintf(stdout, "Successfully parsed GRIB file.\n");
  return 0;
}

#endif
